import numpy as np

import bpy
from blender_plots import blender_utils as bu


class Plot:
    def __init__(self, x, y, z, color=None, name="plot", n_dims=1):
        self.name = name
        self.mesh = bpy.data.meshes.new(self.name)
        self.base_object = bu.new_empty(self.name, self.mesh)
        self.color_material = None
        self._points = None
        self.modifier = self.base_object.modifiers.new(type="NODES", name=name)

        points, self.n_frames, *self.dims = get_points_array(x, y, z, n_dims)
        self.points = points
        self.color = color

    @property
    def n_points(self):
        return np.prod(self.dims)

    @property
    def n_vertices(self):
        return self.n_points * (1 if self.n_frames is None else self.n_frames)

    @property
    def points(self):
        return self._points

    def get_geometry(self):
        raise NotImplementedError

    @points.setter
    def points(self, points):
        if self._points is not None and points.shape != self._points.shape:
            raise ValueError(f"Can't change number of points: was {self._points.shape=}, got {self.points.shape=}")
        self._points = points
        self.update_points()

    def update_points(self):
        if len(self.mesh.vertices) == 0:
            vertices, edges, faces = self.get_geometry()
            self.mesh.from_pydata(vertices, edges, faces)
        elif len(self.mesh.vertices) == len(self._points.reshape(-1, 3)):
            self.mesh.vertices.foreach_set("co", self._points.reshape(-1))
        else:
            raise ValueError(f"Can't change number of vertices,"
                             f"was {len(self.mesh.vertices)=}, got {self._point.shape=}.")

        if self.n_frames is not None:
            bu.set_vertex_attribute(
                self.mesh, bu.Constants.FRAME_INDEX,
                np.arange(0, self.n_frames)[None].repeat(self.n_points, axis=1).reshape(-1)
            )

        self.base_object.data = self.mesh
        self.mesh.update()

    @property
    def color(self):
        return self._color

    @color.setter
    def color(self, color):
        self._color = np.array(color) if color is not None else color
        self.update_color()

    def update_color(self):
        if self._color is not None:
            color, _ = self.tile_data(self._color, [[3], [4]], "color")
            bu.set_vertex_colors(self.mesh, color)
            self.color_material = bu.get_vertex_color_material()
            self.mesh.materials.append(self.color_material)

    def tile_data(self, data_array, valid_dims, name=""):
        """Tile or reshape data_array with shape TxNx(dims), Nx(dims) or (dims) to shape (T*N)x(dims)."""
        if len(self.dims) != 1:
            raise NotImplementedError("Only 1D data can be tiled with base class.")

        match data_array.shape:
            case (self.n_frames, self.n_points, *dims) if dims in valid_dims:
                out_array = data_array.reshape(self.n_vertices, *dims)
            case (self.n_points, *dims) if dims in valid_dims:
                if self.n_frames is not None:
                    out_array = np.tile(data_array, (self.n_frames, *([1] * len(dims))))
                else:
                    out_array = data_array
            case (*dims, ) if dims in valid_dims:
                out_array = np.tile(data_array, (self.n_vertices, *([1]*len(dims))))
            case _:
                raise ValueError(
                    f"Invalid {name} data shape: {data_array.shape} with {self.n_frames=}, {self.n_points=}")
        return out_array, dims


def get_points_array(x, y, z, n_dims=1):
    """Parses x,y,z to a N1xN2x...xN{n_dims}x3 or TxN1xN2x...xN{n_dims}x3 array of points."""
    if (y is None) and (z is None):
        # only x provided, parse it as Nx3 or TxNx3
        x = np.array(x)
        match x.shape:
            case (3, ):
                points = x.reshape(1, 1, 3)
                n_frames = None
                dims = (1 for _ in range(n_dims))
            case (*dims, 3, ) if len(dims) == n_dims:
                points = x
                n_frames = None
            case (n_frames, *dims, 3, ) if len(dims) == n_dims:
                points = x
            case _:
                dims_str = 'x'.join(f'N{i + 1}' for i in range(n_dims))
                raise ValueError(f"Invalid shape for points: {x.shape=}, expected {dims_str}x3 or Tx{dims_str}x3")
    elif (y is not None) and (z is not None):
        # parse x,y,z as N,N,N or TxN,TxN,TxN
        x, y, z = np.array(x), np.array(y), np.array(z)
        match x.shape, y.shape, z.shape:
            case (), (), ():
                points = np.array([x, y, z]).reshape(1, 3)
                n_frames, dims = None, (1 for _ in range(n_dims))
            case (*dims_x, ), (*dims_y, ), (*dims_z, ) if (dims_x == dims_y == dims_z) and len(dims_x) == n_dims:
                points = np.stack([x, y, z], axis=-1)
                n_frames, dims = None, dims_x
            case (tx, *dims_x), (ty, *dims_y), (tz, *dims_z) if (tx == ty == tz) and (dims_x == dims_y == dims_z) and len(dims_x) == n_dims:
                points = np.stack([x, y, z], axis=-1)
                n_frames, dims = tx, dims_x
            case _:
                raise ValueError(f"Incompatible shapes: {x.shape=}, {y.shape=}, {z.shape=}")

    else:
        raise ValueError(f"Eiter both y and z needs to be provided, or neither")
    return points, n_frames, *dims



import numpy as np

import bpy
from blender_plots import plots_base
from blender_plots import blender_utils as bu


class Surface(plots_base.Plot):
    """Create a surface plot.

    Args:
        x,y,z:
            If y and z are not provided: expects x to be a MxNx3 array with xyz positions for points to plot as a surface, or TxMxNx3 for sequence of to animate
            if y and z are provided: expects x,y,z to have shapes MxN or TxMxN arrays for xyz coordinates respectively.
        color: MxNx3 or MxNx4 array or with RGB or RGBA values for each point, or a single RGB/RGBA-value (e.g. (1, 0, 0) for red) to apply to every point.
        name: name to use for blender object. Will delete any previous plot with the same name.
    """

    def __init__(self, x, y=None, z=None, color=None, name="surface"):
        super(Surface, self).__init__(x, y, z, color=color, name=name, n_dims=2)
        if self.n_frames is not None:
            animate(self.modifier, self.n_frames)

    @property
    def n_points_x(self):
        return self.dims[0]

    @property
    def n_points_y(self):
        return self.dims[1]

    def get_geometry(self):
        faces = get_faces(self.n_points_x, self.n_points_y, 1 if self.n_frames is None else self.n_frames)
        return self._points.reshape(-1, 3), [], faces.reshape(-1, 4)

    def tile_data(self, data_array, valid_dims, name=""):
        """Tile or reshape data_array with shape TxMxNx(dims), MxNx(dims) or (dims) to shape (T*M*N)x(dims)."""
        match data_array.shape:
            case (self.n_frames, self.n_points_x, self.n_points_y, *dims) if dims in valid_dims:
                out_array = data_array.reshape(self.n_frames * self.n_points, *dims)
            case (self.n_points_x, self.n_points_y, *dims) if dims in valid_dims:
                if self.n_frames is not None:
                    out_array = np.tile(data_array.reshape(-1, *dims), (self.n_frames, *([1] * len(dims))))
                else:
                    out_array = data_array.reshape(self.n_vertices, *dims)
            case (*dims, ) if dims in valid_dims:
                out_array = np.tile(data_array, (self.n_vertices, *([1]*len(dims))))
            case _:
                raise ValueError(
                    f"Invalid {name} data shape: {data_array.shape} with {self.n_frames=}, {self.n_points=}")
        return out_array, dims


def get_faces(n_points_x, n_points_y, n_frames):
    return np.array([
        [
            j * n_points_x * n_points_y +  np.array([i, i + 1, i + n_points_x + 1, i + n_points_x])
            for i in range(n_points_x * (n_points_y - 1))
            if (i + 1) % n_points_x != 0
        ]
        for j in range(n_frames)
    ])

def animate(base_modifier, n_frames):
    node_linker = bu.get_node_linker(base_modifier)
    visible_geometry = node_linker.new_node(
        "GeometryNodeSeparateGeometry",
        geometry=node_linker.group_input.outputs["Geometry"],
        selection=bu.get_frame_selection_node(base_modifier, n_frames).outputs["Value"]
    ).outputs["Selection"]
    node_linker.new_node("NodeGroupOutput", geometry=visible_geometry)



from dataclasses import dataclass
import numpy as np
import numbers

import bpy
import mathutils as mu
import blender_plots.blender_utils as bu
from blender_plots import plots_base



@dataclass
class Constants:
    MARKER_SCALE = "marker_Scale"
    MARKER_ROTATION = "marker_rotation"
    MARKER_TYPES = {
        "cones": "GeometryNodeMeshCone",
        "cubes": "GeometryNodeMeshCube",
        "cylinders": "GeometryNodeMeshCylinder",
        "grids": "GeometryNodeMeshGrid",
        "ico_spheres": "GeometryNodeMeshIcoSphere",
        "circles": "GeometryNodeMeshCircle",
        "lines": "GeometryNodeMeshLine",
        "uv_spheres": "GeometryNodeMeshUVSphere",
    }


class Scatter(plots_base.Plot):
    """Create a scatterplot.

    Args:
        x,y,z:
            If y and z are not provided: expects x to be a Nx3 array with xyz positions for points to scatter, or TxNx3
                for sequence of scatter plots to animate
            if y and z are provided: expects x,y,z to be length N or TxN arrays for xyz coordinates respectively.
        color: (Tx)Nx3 or (Tx)Nx4 array or with RGB or RGBA values for each point and (optionally) time,
            or a single RGB/RGBA-value (e.g. (1, 0, 0) for red) to apply to every point.
        name: name to use for blender object. Will delete any previous plot with the same name.
        marker_type: select appearance of points. Either MARKER_TYPE, "spheres", bpy_types.Mesh or bpy_types.Object
        marker_rotation: (Tx)Nx3 (euler angles in radians) or (Tx)Nx3x3 (rotation matrices) array specifying the rotation for
            each point and (optionally) time.
        randomize_rotation: If set to True randomize the rotation of each marker. Overrides marker_rotation.
        marker_kwargs: additional arguments for configuring markers
    """

    def __init__(
            self, x, y=None, z=None, color=None, name="scatter",
            marker_type="ico_spheres", marker_scale=None, marker_rotation=None,
            randomize_rotation=False, **marker_kwargs
        ):
        super().__init__(x, y, z, color=color, name=name, n_dims=1)

        if marker_type == "spheres":
            node_linker = add_sphere_markers(self.modifier, n_frames=self.n_frames, **marker_kwargs)
        elif marker_type is not None:
            node_linker = add_mesh_markers(
                self.modifier,
                randomize_rotation=randomize_rotation,
                marker_type=marker_type,
                set_scale=marker_scale is not None,
                n_frames=self.n_frames,
                with_color=color is not None,
                **marker_kwargs
            )
        else:
            node_linker = bu.get_node_linker(self.modifier)
            node_linker.new_input_socket('Point Color', 'NodeSocketMaterial')

        self.modifier[node_linker.input_sockets['Point Color']] = self.color_material
        self.marker_rotation = marker_rotation
        self.marker_scale = marker_scale
        self.base_object.data.update()

    def get_geometry(self):
        return self._points.reshape(-1, 3), [], []

    @property
    def marker_scale(self):
        return self._marker_scale

    @marker_scale.setter
    def marker_scale(self, marker_scale):
        self._marker_scale = np.array(marker_scale) if marker_scale is not None else marker_scale
        self.update_marker_scale()

    def update_marker_scale(self):
        if self._marker_scale is not None:
            marker_scale, marker_dims = self.tile_data(self._marker_scale, [[3], []], "marker scale")
            if marker_dims == []:
                marker_scale = np.array([marker_scale] * 3).T
            bu.set_vertex_attribute(self.mesh, Constants.MARKER_SCALE, marker_scale, "FLOAT_VECTOR")

    @property
    def marker_rotation(self):
        return self._marker_rotation

    @marker_rotation.setter
    def marker_rotation(self, marker_rotation):
        self._marker_rotation = np.array(marker_rotation) if marker_rotation is not None else marker_rotation
        self.update_marker_rotation()

    def update_marker_rotation(self):
        if self._marker_rotation is not None:
            marker_rotation, rotation_dims = self.tile_data(self._marker_rotation, [[3], [3, 3]], "marker rotation")
            if rotation_dims == [3, 3]:
                marker_rotation = np.stack([np.array(mu.Matrix(r).to_euler()) for r in marker_rotation])
            bu.set_vertex_attribute(self.mesh, Constants.MARKER_ROTATION, marker_rotation, "FLOAT_VECTOR")


def add_mesh_markers(base_modifier, marker_type, randomize_rotation=False, set_scale=False,
                     n_frames=0, with_color=False, **marker_kwargs):
    """Create a geometry node modifier that instances a mesh on each vertex.
    Args:
        base_modifier: modifier to add markers to.
        marker_type: name of marker type (see MARKER_TYPES), or a blender mesh/object to use as marker
        randomize_rotation: if True each mesh instance will be given a random rotation (uniform euler angles)
        set_scale: if True use the MARKER_SCALE attribute to set the marker scale
        n_frames: number of frames to animate, no animation if set to 0.
        marker_kwargs: additional arguments for configuring markers
    """
    node_linker = bu.get_node_linker(base_modifier)
    node_linker.new_input_socket('Point Color', 'NodeSocketMaterial')

    points_socket = node_linker.new_node(
        "GeometryNodeMeshToPoints",
        mesh=node_linker.group_input.outputs["Geometry"]
    ).outputs["Points"]

    if marker_type in Constants.MARKER_TYPES:
        # use one of the default marker types
        mesh_socket = node_linker.new_node(node_type=Constants.MARKER_TYPES[marker_type], **marker_kwargs).outputs["Mesh"]
    elif isinstance(marker_type, (bpy.types.Object, bpy.types.Collection)):
        # use custom object or collection as marker
        subtype = marker_type.__class__.__name__
        node_linker.new_input_socket('Point Instance', f'NodeSocket{subtype}')
        base_modifier[node_linker.input_sockets['Point Instance']] = marker_type
        mesh_socket = node_linker.new_node(
            f"GeometryNode{subtype}Info",
            **{subtype: node_linker.group_input.outputs["Point Instance"]}
        ).outputs[{'Object': 'Geometry', 'Collection': 'Instances'}[subtype]]
        marker_type.hide_viewport = True
        marker_type.hide_render = True
    else:
        raise TypeError(f"Invalid marker type: {marker_type}, expected Object or Collection, "
                        f"or one of: {', '.join(Constants.MARKER_TYPES)}")

    if with_color:
        mesh_socket = node_linker.new_node(
            "GeometryNodeSetMaterial",
            geometry=mesh_socket,
            material=node_linker.group_input.outputs["Point Color"]
        ).outputs["Geometry"]

    marker_scale = node_linker.new_node(
        "GeometryNodeInputNamedAttribute",
        data_type="FLOAT_VECTOR",
        name=Constants.MARKER_SCALE,
    )
    marker_rotation = node_linker.new_node(
        "GeometryNodeInputNamedAttribute",
        data_type="FLOAT_VECTOR",
        name=Constants.MARKER_ROTATION,
    )
    instance_on_points_node = node_linker.new_node(
        "GeometryNodeInstanceOnPoints",
        points=points_socket,
        selection=None if n_frames is None else bu.get_frame_selection_node(base_modifier, n_frames).outputs["Value"],
        instance=mesh_socket,
        rotation=marker_rotation.outputs["Attribute"],
        scale=marker_scale.outputs["Attribute"] if set_scale else [1, 1, 1],
    )
    if randomize_rotation:
        # these rotation are not uniform (some orientations will be more likely than others)
        # but it usually looks decent
        random_euler = node_linker.new_node("FunctionNodeRandomValue")
        random_euler.data_type = "FLOAT_VECTOR"
        random_euler.inputs['Max'].default_value = (180, 180, 180)
        node_linker.link(random_euler.outputs["Value"], instance_on_points_node.inputs["Rotation"])

    realize_instances_node = node_linker.new_node(
        "GeometryNodeRealizeInstances",
        geometry=instance_on_points_node.outputs["Instances"]
    )
    node_linker.new_node("NodeGroupOutput", geometry=realize_instances_node.outputs["Geometry"])
    return node_linker


def add_sphere_markers(base_modifier, n_frames, **marker_kwargs):
    """Create a geometry node modifier that adds a point on each vertex. This will result in perfect spheres, only
        visible in rendered view with rendering engine set to `Cycles`
    Args:
        base_modifier: modifier to add sphere markers to.
        n_frames: number of frames to animate, no animation if set to 0.
        marker_kwargs: arguments to passed to node_linker.new_node when generating point node. e.g. radius=0.1
    """
    node_linker = bu.get_node_linker(base_modifier)
    node_linker.new_input_socket('Point Color', 'NodeSocketMaterial')

    points = node_linker.new_node(
        "GeometryNodeMeshToPoints",
        mesh=node_linker.group_input.outputs["Geometry"],
        selection=None if n_frames is None else bu.get_frame_selection_node(base_modifier, n_frames).outputs["Value"],
        **marker_kwargs,
    ).outputs["Points"]
    node = node_linker.new_node(
        "GeometryNodeSetMaterial",
        geometry=points,
        material=node_linker.group_input.outputs["Point Color"]
    )
    node_linker.new_node("NodeGroupOutput", geometry=node.outputs["Geometry"])
    return node_linker



from blender_plots import blender_utils, plots_base, scatter, surface, arrows, marker_utils
from blender_plots.scatter import Scatter
from blender_plots.surface import Surface
from blender_plots.arrows import Arrows



from blender_plots import plots_base, scatter
from blender_plots import blender_utils as bu

import numpy as np

class Arrows(plots_base.Plot):
    """Create an arrow plot.

    Args:
        start: xyz coordinate(s) of arrow starts (3, Nx3 or TxNx3 array - see scatter.py)
        vector: xyz components of arrow vectors, same size as `start`
        end: xyz coordinate(s) of arrow ends (can be provided instead of `vector`)
        color: (Tx)Nx3 or (Tx)Nx4 array or with RGB or RGBA values for each point and (optionally) time,
            or a single RGB/RGBA-value (e.g. (1, 0, 0) for red) to apply to every point.
        name: name to use for blender object. Will delete any previous plot with the same name.
        head_length: length of the arrow head.
        radius: radius of the arrow stem.
        radius_ratio: ratio of the arrow head radius to the stem radius.
        end_trim_length: length by which to reduce each arrow. If set to 0 the arrow tip will be exactly at `start`+`vector` (or `end` if provided)
    """

    def __init__(self, start, vector=None, end=None, color=None, name='arrows', head_length=1, radius=1, radius_ratio=1.5, end_trim_length=0):
        super(Arrows, self).__init__(start, None, None, color=color, name=name, n_dims=1)

        self.node_linker = add_arrows(self.modifier, self.n_frames, head_length, radius, radius_ratio, end_trim_length)

        self.modifier[self.node_linker.input_sockets['Point Color']] = self.color_material

        if (end is not None) and (vector is not None):
            raise ValueError("Cannot provide both 'end' and 'vector' as input arguments.")
        elif end is not None:
            self.arrows = end - start
        else:
            self.arrows = vector
        self.base_object.data.update()

    def get_geometry(self):
        return self._points.reshape(-1, 3), [], []

    @property
    def arrows(self):
        return self._arrows

    @arrows.setter
    def arrows(self, arrows):
        self._arrows = np.array(arrows) if arrows is not None else arrows
        self.update_arrows()

    def update_arrows(self):
        if self._arrows is not None:
            arrows, marker_dims = self.tile_data(self._arrows, [[3], []], "marker scale")
            if marker_dims == []:
                arrows = np.array([arrows] * 3).T
            bu.set_vertex_attribute(self.mesh, bu.Constants.ARROWS, arrows, "FLOAT_VECTOR")

def add_arrows(base_modifier, n_frames, head_length, radius, radius_ratio, end_trim_length):
    """
    """
    node_linker = bu.get_node_linker(base_modifier)
    node_linker.new_input_socket('Point Color', 'NodeSocketMaterial')
    if n_frames is None:
        frame_selection = None
    else:
        frame_selection = bu.get_frame_selection_node(base_modifier, n_frames).outputs["Value"]

    points_socket = node_linker.new_node(
        "GeometryNodeMeshToPoints",
        mesh=node_linker.group_input.outputs["Geometry"]
    ).outputs["Points"]

    arrows = node_linker.new_node(
        "GeometryNodeInputNamedAttribute", data_type="FLOAT_VECTOR", name=bu.Constants.ARROWS
    ).outputs["Attribute"]
    lengths = node_linker.new_node("ShaderNodeVectorMath", operation="LENGTH", vector=arrows).outputs["Value"]
    lengths = node_linker.new_node("ShaderNodeMath", operation="SUBTRACT", input_0=lengths, input_1=head_length+end_trim_length).outputs["Value"]
    rotations = node_linker.new_node("FunctionNodeAlignEulerToVector", axis="Z", vector=arrows).outputs["Rotation"]

    cylinder_mesh = node_linker.new_node("GeometryNodeMeshCylinder", depth=1, radius=radius).outputs["Mesh"]
    cylinder_mesh = node_linker.new_node("GeometryNodeTransform", geometry=cylinder_mesh, translation=[0, 0, .5]).outputs["Geometry"]
    colored_cylinder = node_linker.new_node(
        "GeometryNodeSetMaterial",
        geometry=cylinder_mesh,
        material=node_linker.group_input.outputs["Point Color"]
    ).outputs["Geometry"]
    cylinder_instances = node_linker.new_node(
        "GeometryNodeInstanceOnPoints",
        points=points_socket,
        selection=frame_selection,
        instance=colored_cylinder,
        rotation=rotations,
        scale=node_linker.new_node("ShaderNodeCombineXYZ", x=1, y=1, z=lengths).outputs["Vector"]
    ).outputs["Instances"]
    cylinder_geometry = node_linker.new_node("GeometryNodeRealizeInstances", geometry=cylinder_instances).outputs["Geometry"]

    cone_mesh = node_linker.new_node("GeometryNodeMeshCone", depth=head_length, radius_bottom=radius * radius_ratio).outputs["Mesh"]
    colored_cone = node_linker.new_node(
        "GeometryNodeSetMaterial",
        geometry=cone_mesh,
        material=node_linker.group_input.outputs["Point Color"]
    ).outputs["Geometry"]
    cone_instances = node_linker.new_node(
        "GeometryNodeInstanceOnPoints",
        points=points_socket,
        selection=frame_selection,
        instance=colored_cone,
        rotation=rotations,
    ).outputs["Instances"]
    cone_instances = node_linker.new_node(
        "GeometryNodeTranslateInstances",
        instances=cone_instances,
        translation=node_linker.new_node("ShaderNodeCombineXYZ", x=0, y=0, z=lengths).outputs["Vector"],
    ).outputs["Instances"]
    cone_geometry = node_linker.new_node("GeometryNodeRealizeInstances", geometry=cone_instances).outputs["Geometry"]
    arrow_geometry_node = node_linker.new_node("GeometryNodeJoinGeometry", geometry=cylinder_geometry)
    node_linker.node_group.links.new(cone_geometry, arrow_geometry_node.inputs[0])

    node_linker.new_node("NodeGroupOutput", geometry=arrow_geometry_node.outputs["Geometry"])
    return node_linker



from dataclasses import dataclass

import bpy
import numpy as np


@dataclass
class Constants:
    MARKER_COLOR = "marker_color"
    FRAME_INDEX = "frame_index"
    ARROWS = "arrows"


class NodeLinker:
    """Wrapper for bpy.types.GeometryNodeTree which simplifies creating large node trees."""

    def __init__(self, node_group):
        self.node_group = node_group
        # blender creates socket names based on order they are created, so we keep a mapping from user-provided socket name to internal name
        self.input_sockets = {}

    def new_node(self, node_type, **kwargs):
        """Adds a new node to the node group

        Args:
            node_type: type of node to add (e.g. GeometryNodeInstanceOnPoints) - to find available names you can either
                hover over the node name when adding one through the node editor, or check the `bl_idname`
                property of a node in python.
            kwargs: arguments in one of the following forms:
                input_socket_name=node_socket: will connect the supplied node_socket to the input socket
                    (e.g. mesh=node_group["Group Input"].outputs["Geometry"] for GeometryNodeMeshToPoints)
                input_socket_name=value: set the value of input socket (e.g. size=(5, 1, 1) for GeometryNodeMeshCube).
                input_i=value: set input i of node to value, useful in case multiple inputs have the same name.
        """
        if node_type == "NodeGroupOutput":
            # can only have one group output
            node = self.node_group.nodes["Group Output"]
        else:
            node = self.node_group.nodes.new(node_type)

        for key, value in kwargs.items():
            if value is not None:
                match key.split("_"):
                    case ["input", i] if i.isdigit():
                        blender_key = int(i)
                    case _:
                        blender_key = python_arg_to_blender_key(key)
                if isinstance(value, bpy.types.NodeSocket):
                    self.link(value, node.inputs[blender_key])
                elif (isinstance(blender_key, int) or blender_key in node.inputs) and hasattr(node.inputs[blender_key], "default_value"):
                    node.inputs[blender_key].default_value = value # TODO: make sure this is done first, as it resets everything else
                elif hasattr(node, key):
                    setattr(node, key, value)
                else:
                    raise ValueError(f"Node {node} has no attribute {key} or input {blender_key}.")
        return node

    def link(self, from_socket, to_socket):
        self.node_group.links.new(from_socket, to_socket)

    def new_input(self, input_type, input_name):
        self.node_group.inputs.new(input_type, input_name)

    def new_input_socket(self, name, socket_type):
        if bpy.app.version[0] >= 4: # node group input/output interface changed in 4.0
            self.node_group.interface.new_socket(name, socket_type=socket_type)
        else:
            self.node_group.inputs.new(socket_type, name)
        socket_index = len(self.input_sockets) + 2 # starts at 2
        self.input_sockets[name] = socket_input_key(socket_index)

    @property
    def group_input(self):
        """Add input which can be accessed through modifiers panel."""
        return self.node_group.nodes["Group Input"]


def delete(obj, with_children=False):
    """Delete blender object and its children"""
    if with_children:
        for child in obj.children:
            delete(child, with_children=True)
    if isinstance(obj, bpy.types.Object):
        bpy.data.objects.remove(obj, do_unlink=True)
    elif isinstance(obj, bpy.types.Collection):
        bpy.data.collections.remove(obj, do_unlink=True)
    else:
        raise ValueError(f'Failed to delete object {obj}: unrecognized type')

def new_collection(name):
    if name in bpy.data.collections:
        delete(bpy.data.collections[name], with_children=True)
    collection = bpy.data.collections.new(name)
    bpy.context.collection.children.link(collection)
    return collection

def new_empty(name, object_data=None, select=True, collection=None, location=None, rotation_euler=None):
    """Create new empty blender object with specified name and data, deletes any previous object with the same name."""
    if name in bpy.data.objects:
        delete(bpy.data.objects[name], with_children=True)

    new_object = bpy.data.objects.new(name, object_data)
    if collection is None:
        collection = bpy.context.collection
    collection.objects.link(new_object)

    if select:
        bpy.context.view_layer.objects.active = new_object
    if location is not None:
        new_object.location = location
    if rotation_euler is not None:
        new_object.rotation_euler = rotation_euler
    return new_object


def add_modifier(base_object, modifier_type, **kwargs):
    modifier = base_object.modifiers.new(name=modifier_type.lower(), type=modifier_type)
    for key, value in kwargs.items():
        setattr(modifier, key, value)
    return modifier


def set_vertex_attribute(mesh, attribute_name,  attribute_values, attribute_type="FLOAT"):
    if attribute_name not in mesh.attributes:
        mesh.attributes.new(name=attribute_name, type=attribute_type, domain="POINT")
    data_type = "vector" if attribute_type == "FLOAT_VECTOR" else "value"
    mesh.attributes[attribute_name].data.foreach_set(data_type, attribute_values.reshape(-1))


def python_arg_to_blender_key(arg):
    """convert python argument to geometry node name, e.g. radius->Radius, instance_index->Instance Index"""
    return ' '.join([s.capitalize() for s in arg.split('_')])

def socket_input_key(i):
    return ('Socket_' if bpy.app.version[0] >= 4 else 'Input_') + f'{i}'

# From https://developer.blender.org/diffusion/B/browse/master/release/scripts/startup/bl_operators/geometry_nodes.py$7
def get_node_linker(modifier):
    if modifier.node_group is not None:
        return NodeLinker(modifier.node_group)
    group = bpy.data.node_groups.new("Geometry Nodes", 'GeometryNodeTree')
    if bpy.app.version[0] >= 4: # node group input/output interface changed in 4.0
        input_node = group.nodes.new('NodeGroupInput')
        output_node = group.nodes.new('NodeGroupOutput')
        group.interface.new_socket('Geometry', in_out='INPUT', socket_type='NodeSocketGeometry')
        group.interface.new_socket('Geometry', in_out='OUTPUT', socket_type='NodeSocketGeometry')
    else:
        """Create new node group, useful in blender 3.2 since node group is not added to node modifiers by default."""
        group.inputs.new('NodeSocketGeometry', "Geometry")
        group.outputs.new('NodeSocketGeometry', "Geometry")
        input_node = group.nodes.new('NodeGroupInput')
        output_node = group.nodes.new('NodeGroupOutput')
    output_node.is_active_output = True

    input_node.select = False
    output_node.select = False

    input_node.location.x = -200 - input_node.width
    output_node.location.x = 200

    group.links.new(output_node.inputs[0], input_node.outputs[0])
    modifier.node_group = group
    return NodeLinker(modifier.node_group)

def set_vertex_colors(mesh, color):
    """Add a marker_color attribute to each vertex in `mesh` with values from (n_vertices)x(3 or 4) array `color`"""
    if color.shape[1] == 3:
        color = np.hstack([color, np.ones((len(color), 1))])
    elif not color.shape[1] == 4:
        raise ValueError(f"Invalid color array shape {color.shape}, expected Nx3 or Nx4")
    if len(mesh.vertices) != len(color):
        raise ValueError(f"Got {len(mesh.vertices)} vertices and {len(color)} color values")

    if Constants.MARKER_COLOR not in mesh.attributes:
        mesh.attributes.new(name=Constants.MARKER_COLOR, type="FLOAT_COLOR", domain="POINT")
    mesh.attributes[Constants.MARKER_COLOR].data.foreach_set("color", color.reshape(-1))


def get_vertex_color_material():
    """Create a material that obtains its color from the marker_color attribute"""
    material = bpy.data.materials.new("color")
    material.use_nodes = True
    bsdf = material.node_tree.nodes.get("Principled BSDF")

    color_node = material.node_tree.nodes.new("ShaderNodeAttribute")
    color_node.attribute_name = Constants.MARKER_COLOR

    material.node_tree.links.new(color_node.outputs["Color"], bsdf.inputs["Base Color"])
    material.node_tree.links.new(color_node.outputs["Alpha"], bsdf.inputs["Alpha"])

    material.blend_method = "BLEND"
    return material

def add_mesh_color(mesh, color):
    """Add uniform color to mesh."""
    if len(color) == 3:
        color = (*color, 1)
    material = bpy.data.materials.new("color")
    material.diffuse_color = color
    mesh.materials.append(material)

def get_frame_selection_node(modifier, n_frames):
    """Add node that filters points based on the Frame Index property."""
    node_linker = NodeLinker(modifier.node_group)
    frame_index = node_linker.new_node(
        "GeometryNodeInputNamedAttribute",
        data_type="FLOAT",
        name=Constants.FRAME_INDEX,
    )
    frame_selection_node = node_linker.new_node(
        "ShaderNodeMath",
        operation="COMPARE",
        input_1=frame_index.outputs[1],
        input_2=0.5
    )

    # link compare node to frame index
    action = bpy.data.actions.new("AnimationAction")
    fcurve = action.fcurves.new(data_path=f'nodes["{frame_selection_node.name}"].inputs[0].default_value', index=0)
    fcurve.keyframe_points.add(2)
    fcurve.keyframe_points.foreach_set("co", [0, 0, n_frames, n_frames])
    bpy.context.scene.frame_end = n_frames - 1

    modifier.node_group.animation_data_create()
    modifier.node_group.animation_data.action = action

    return frame_selection_node

def new_mesh(vertices, faces, name='mesh', color=None, location=None, rotation_euler=None):
    """Create mesh with uniform color."""
    bmesh = bpy.data.meshes.new(name=name)
    bmesh.from_pydata(vertices, [], faces)
    mesh_object = new_empty(
        name,
        bmesh,
        location=location,
        rotation_euler=rotation_euler,
    )

    if color is not None:
        material = bpy.data.materials.new(name)
        material.use_nodes = True
        bsdf = material.node_tree.nodes['Principled BSDF']
        bsdf.inputs['Base Color'].default_value = color
        mesh_object.data.materials.append(material)
    return mesh_object
